import random
from math import sqrt
import io
import copy

try:
    from tqdm import tqdm
except:
    tqdm = lambda x: x


class Test:

    MAX_N = 1000000
    MAX_M = 1000000
    MAX_A = 1000
    MAX_D = MAX_A - 1

    def __init__(self):
        self.n = 0
        self.edges = []

    def add_edge(self, u, v, d):
        self.n = max(self.n, u + 1, v + 1)
        assert -Test.MAX_D <= d <= Test.MAX_D
        self.edges.append((u, v, d))

    def print_to_file(self, f, shuffled=True):
        f.write("{} {}\n".format(self.n, len(self.edges)))
        p = list(range(self.n))
        if shuffled:
            random.shuffle(p)
            random.shuffle(self.edges)

        for u, v, d in self.edges:
            a, b = 0, d
            a += 1 - min(0, d)
            b += 1 - min(0, d)
            if shuffled:
                shift = random.randint(0, Test.MAX_A - max(a, b))
                a += shift
                b += shift
            assert 1 <= a <= Test.MAX_A
            assert 1 <= b <= Test.MAX_A
            f.write("{} {} {} {}\n".format(p[u] + 1, p[v] + 1, a, b))

    def __repr__(self):
        f = io.StringIO()
        self.print_to_file(f, shuffled=False)
        res = f.getvalue()
        f.close()
        return res

    def break_edge(self, edge_index):
        u, v, old_d = self.edges[edge_index]
        new_d = old_d
        while new_d == old_d:
            new_d = random_d()
        self.edges[edge_index] = (u, v, new_d)

    def break_random_edge(self):
        ei = random.randrange(len(self.edges))
        self.break_edge(ei)

    def check(self, maxn, maxm, limited):
        if self.n > maxn:
            raise ValueError("Too many vertices: {}".format(self.maxn))
        if len(self.edges) > maxm:
            raise ValueError("Too many edges: {}".format(len(self.edges)))

        deg = [0] * self.n
        for u, v, _ in self.edges:
            if u == v:
                raise ValueError("Self-loop: {}".format(u))
            deg[u] += 1
            deg[v] += 1

        if limited:
            big_cnt = 0
            for u in range(self.n):
                if deg[u] > 2:
                    big_cnt += 1

            if big_cnt > 6:
                raise ValueError("Too many vertices of degree 3 or more.")

        return True


def random_almost(n):
    return max(1, random.randint(int(0.8 * n), n))


def random_d():
    return random.randint(-Test.MAX_D, Test.MAX_D)


def random_edge(n):
    u = random.randrange(n)
    v = random.randrange(n)
    while v == u:
        v = random.randrange(n)
    return u, v


def gen_long_loop(n):
    t = Test()

    d_half_1 = [random_d() for _ in range(n // 2)]
    d_half_2 = [-x for x in d_half_1]
    ds = d_half_1 + d_half_2
    if n % 2 == 1:
        ds.append(0)
    random.shuffle(ds)

    for i, d in enumerate(ds):
        t.add_edge(i, (i + 1) % n, d)

    return t


def gen_short_loops(n):
    # Pairs of vertices connected in loops
    t = Test()
    p = list(range(n))
    random.shuffle(p)
    for u, v in zip(p[::2], p[1::2]):
        d = random_d()
        t.add_edge(u, v, d)
        t.add_edge(v, u, -d)

    return t


def gen_joined_short_loops(n):
    # Pairs of vertices connected in loops, the loops being connected one-directionally
    t = Test()
    for u in range(0, n, 2):
        v = u + 1
        d = random_d()
        t.add_edge(u, v, d)
        t.add_edge(v, u, -d)
        if u:
            t.add_edge(u, u - 2, random_d())

    return t


def gen_multiedges(_n, m):  # n is ignored but is here for consistency
    assert m >= 2

    t = Test()
    d = random_d()
    t.add_edge(0, 1, d)
    t.add_edge(1, 0, -d)
    left = random.randint(0, m - 2)
    for i in range(m - 2):
        if i < left:
            t.add_edge(0, 1, d)
        else:
            t.add_edge(1, 0, -d)

    return t


def gen_complete(n):
    t = Test()
    hs = [random.randint(0, Test.MAX_D) for _ in range(n)]
    for u in range(n):
        for v in range(n):
            if u != v:
                t.add_edge(u, v, hs[u] - hs[v])

    return t


def gen_flower(n):
    # One central vertex, others are connected to it with a loop. Loops are small.
    # At most 2N edges
    last = 0
    t = Test()
    hs = [random.randint(0, Test.MAX_D) for _ in range(n)]
    for u in range(1, n):
        t.add_edge(last, u, hs[last] - hs[u])
        if random.random() < 0.5:
            last = 0
            t.add_edge(u, 0, hs[u] - hs[0])
        else:
            last = u

    if last != 0:
        t.add_edge(n - 1, 0, hs[n - 1] - hs[0])

    return t


def gen_random(n, m, bidirectional=False):
    t = Test()
    hs = [random.randint(0, Test.MAX_D) for _ in range(n)]

    if bidirectional:
        m = m // 2

    for _ in range(m):
        u, v = random_edge(n)
        t.add_edge(u, v, hs[u] - hs[v])
        if bidirectional:
            t.add_edge(v, u, hs[v] - hs[u])

    return t


def gen_grid(n, m):
    t = Test()
    n = min(n, m // 2)

    sz = int(sqrt(n))

    def pos_to_index(x, y):
        return x * sz + y

    for i in range(sz):
        for j in range(sz):
            if i > 0:
                t.add_edge(pos_to_index(i, j), pos_to_index(i - 1, j), random_d())
            if j > 0:
                t.add_edge(pos_to_index(i, j), pos_to_index(i, j - 1), random_d())

    return t


def augment(test, maxm, edges_to_add):
    # Added edge is likely to be a bad edge
    test = copy.deepcopy(test)
    for i in range(edges_to_add):
        if len(test.edges) == maxm:
            break
        u, v = random_edge(test.n)
        test.add_edge(u, v, random_d())

    return test


def gen_all(path):
    groups = [
        (20, 20, False, False),
        (Test.MAX_N, 1000, True, False),
        (Test.MAX_N, Test.MAX_M, True, False),
        (1000, 1000, False, True),
        (1000, 1000, False, False),
        (Test.MAX_N, Test.MAX_M, False, True),
        (Test.MAX_N, Test.MAX_M, False, False),
    ]

    for group_i, (maxn, maxm, limited, connected) in enumerate(tqdm(groups)):
        good_tests = [
            gen_long_loop(random_almost(min(maxn, maxm))),
            gen_flower(random_almost(min(maxn, maxm // 2))),
            gen_multiedges(0, random_almost(maxm)),
        ]

        if not connected:
            good_tests.extend([
                gen_joined_short_loops(8),  # low number of vertices; to be ok in limited as well
                gen_short_loops(random_almost(min(maxn, maxm))),
            ])

        if not limited:
            good_tests.extend([
                gen_complete(random_almost(min(maxn, int(sqrt(maxm))))),
            ])
            if not connected:
                good_tests.extend([
                    gen_random(random_almost(maxn), random_almost(maxm)),
                    gen_random(random_almost(maxm // 2), random_almost(maxm)),
                ])

        bad_tests = []

        for test_i, test in enumerate(good_tests):
            if group_i in [1, 2, 4, 6]:
                # V techto sadach chceme min testu; misto duplikace nahrazujeme
                if test_i % 2 == 0:
                    test.break_random_edge()
            else:
                test2 = copy.deepcopy(test)
                test2.break_random_edge()
                bad_tests.append(test2)

        # if not limited:
        #     for t in good_tests:
        #         bad_tests.append(augment(t, maxm, 1))

        # Priklady nejsou vesmes souvisle
        if connected:
            all_tests = good_tests + bad_tests
        else:
            all_tests = fixed_tests + good_tests + bad_tests

        if group_i == 4:
            all_tests.append(gen_grid(random_almost(maxn), random_almost(maxm)))
        if group_i == 6:
            all_tests.append(gen_grid(random_almost(maxn), random_almost(maxm)))
            all_tests.append(gen_grid(random_almost(maxn), random_almost(maxm)))

        group_mapping = [
            [1, 2, 3],
            [4],
            [5],
            [6],
            [7],
            [8],
            [9, 10]
        ]

        for test_i, test in enumerate(all_tests):
            # Je vice sad se stejnymi limity, rozdelime testy mezi ne
            target_group = group_mapping[group_i][test_i % len(group_mapping[group_i])]

            with open("{}/{:02}_{:02}.in".format(path, target_group, test_i + 1), "w") as f:
                if isinstance(test, str):  # Fixed test
                    for line in test.split("\n"):
                        if line.strip() != "":
                            f.write("{}\n".format(line.strip()))
                else:
                    try:
                        test.check(maxn, maxm, limited)
                    except ValueError as e:
                        print("Error in test {}_{}".format(group_i + 1, test_i + 1))
                        print(e)
                        return

                    test.print_to_file(f)

fixed_tests = [
    """
    4 6
    1 2 1 1
    2 3 1 1
    3 1 1 1
    2 4 1 1
    4 1 1 2
    4 1 1 1
    """,
    """
    3 3
    1 2 1 1
    2 3 1 2
    3 1 2 1
    """,
    """
    3 2
    1 2 1 1
    2 3 1 2
    """,
    """
    2 1
    1 2 1 1000
    """,
    """
    2 1
    2 1 1 1000
    """,
    """
    20 2
    7 13 1 1000
    13 7 1 1
    """,
]


if __name__ == '__main__':
    random.seed(1819)
    gen_all("./p1-2data/")
